"""Gumbo HTML parser.

These are the Python bindings for Gumbo.  All public API classes and functions
are exported from this module.  They include:

- CTypes representations of all structs and enums defined in gumbo.h.  The
  naming convention is to take the C name and strip off the "Gumbo" prefix.

- A low-level wrapper around the gumbo_parse function, returning the classes
  exposed above.  Usage:

  import gumbo
  with gumboc.parse(text, **options) as output:
    do_stuff_with_doctype(output.document)
    do_stuff_with_parse_tree(output.root)

- Higher-level bindings that mimic the API provided by html5lib.  Usage:

  from gumbo import html5lib

  This requires that html5lib be installed (it uses their treebuilders), and is
  intended as a drop-in replacement.

- Similarly, higher-level bindings that mimic BeautifulSoup and return
  BeautifulSoup objects.  For this, use:

  import gumbo
  soup = gumbo.soup_parse(text, **options)

  It will give you back a soup object like BeautifulSoup.BeautifulSoup(text).
"""

from gumbo.gumboc import *

try:
  from gumbo import html5lib_adapter as html5lib
except ImportError:
  # html5lib not installed
  pass

try:
  from gumbo.soup_adapter import parse as soup_parse
except ImportError:
  # BeautifulSoup not installed
  pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=2:sw=2:softtabstop=2:smarttab:expandtab

from __future__ import unicode_literals, print_function

# Copyright 2012 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys

PY3 = sys.version_info[0] >= 3

if PY3:
    text_type = str
    binary_type = bytes
else:
    text_type = unicode
    binary_type = str

# When supporting both python 2 and 3 using one code base, using str(obj) is confusing 
# at best since its return type is python version specific
# Notes: 
#   - The unicode(obj) operator does not exist in PY3
#   - The bytes(obj) exists and works on both python >= 2.6


"""CTypes bindings for the Gumbo HTML5 parser.

This exports the raw interface of the library as a set of very thin ctypes
wrappers.  It's intended to be wrapped by other libraries to provide a more
Pythonic API.
"""

__author__ = 'jdtang@google.com (Jonathan Tang)'

import contextlib
import ctypes
import os.path
import gumboc_tags

_name_of_lib = 'libgumbo.so'
if sys.platform.startswith('darwin'):
  _name_of_lib = 'libgumbo.dylib'
elif sys.platform.startswith('win'):
  _name_of_lib = "gumbo.dll"

try:
  # First look for a freshly-built .so in the .libs directory, for development.
  _dll = ctypes.cdll.LoadLibrary(os.path.join(
      os.path.dirname(__file__), '..', '..', '.libs', _name_of_lib))
except OSError:
  # PyPI or setuptools install, look in the current directory.
  _dll = ctypes.cdll.LoadLibrary(os.path.join(
      os.path.dirname(__file__), _name_of_lib))
except OSError:
  # System library, on unix or mac osx
  _dll = ctypes.cdll.LoadLibrary(_name_of_lib)

# Some aliases for common types.
_bitvector = ctypes.c_uint
_Ptr = ctypes.POINTER

class EnumMetaclass(type(ctypes.c_uint)):
  def __new__(metaclass, name, bases, cls_dict):
    cls = type(ctypes.c_uint).__new__(metaclass, name, bases, cls_dict)
    if name == 'Enum':
      return cls
    try:
      for i, value in enumerate(cls_dict['_values_']):
        setattr(cls, value, cls.from_param(i))
    except KeyError:
      raise ValueError('No _values_ list found inside enum type.')
    except TypeError:
      raise ValueError('_values_ must be a list of names of enum constants.')
    return cls

def with_metaclass(mcls):
    def decorator(cls):
        body = vars(cls).copy()
        # clean out class body
        body.pop('__dict__', None)
        body.pop('__weakref__', None)
        return mcls(cls.__name__, cls.__bases__, body)
    return decorator

@with_metaclass(EnumMetaclass)
class Enum(ctypes.c_uint):
  @classmethod
  def from_param(cls, param):
    if isinstance(param, Enum):
      if param.__class__ != cls:
        raise ValueError("Can't mix enums of different types")
      return param
    if param < 0 or param > len(cls._values_):
      raise ValueError('%d is out of range for enum type %s; max %d.' %
                       (param, cls.__name__, len(cls._values_)))
    return cls(param)

  def __eq__(self, other):
    return self.value == other.value

  def __ne__(self, other):
    return self.value != other.value

  def __hash__(self):
    return hash(self.value)

  def __repr__(self):
    try:
      return self._values_[self.value]
    except IndexError:
      raise IndexError('Value %d is out of range for %r' %
                       (self.value, self._values_))



class StringPiece(ctypes.Structure):
  _fields_ = [
      ('data', _Ptr(ctypes.c_char)),
      ('length', ctypes.c_size_t),
      ]

  def __len__(self):
    if PY3:
      return len(ctypes.string_at(self.data, self.length).decode('utf-8'))
    return self.length

  def __str__(self):
    # Note in Python3 the str() method may **never** return bytes 
    if PY3:
      return ctypes.string_at(self.data, self.length).decode('utf-8')
    return ctypes.string_at(self.data, self.length)

  def __bytes__(self):
    return ctypes.string_at(self.data, self.length)


class SourcePosition(ctypes.Structure):
  _fields_ = [
      ('line', ctypes.c_uint),
      ('column', ctypes.c_uint),
      ('offset', ctypes.c_uint)
      ]
SourcePosition.EMPTY = SourcePosition.in_dll(_dll, 'kGumboEmptySourcePosition')


class AttributeNamespace(Enum):
  URLS = [
      'http://www.w3.org/1999/xhtml',
      'http://www.w3.org/1999/xlink',
      'http://www.w3.org/XML/1998/namespace',
      'http://www.w3.org/2000/xmlns',
  ]
  _values_ = ['NONE', 'XLINK', 'XML', 'XMLNS']

  def to_url(self):
    return self.URLS[self.value]


class Attribute(ctypes.Structure):
  _fields_ = [
      ('namespace', AttributeNamespace),
      ('name', ctypes.c_char_p),
      ('original_name', StringPiece),
      ('value', ctypes.c_char_p),
      ('original_value', StringPiece),
      ('name_start', SourcePosition),
      ('name_end', SourcePosition),
      ('value_start', SourcePosition),
      ('value_end', SourcePosition)
      ]


class Vector(ctypes.Structure):
  _type_ = ctypes.c_void_p
  _fields_ = [
      ('data', _Ptr(ctypes.c_void_p)),
      ('length', ctypes.c_uint),
      ('capacity', ctypes.c_uint)
      ]

  class Iter(object):
    def __init__(self, vector):
      self.current = 0
      self.vector = vector

    def __iter__(self):
      return self

    def __next__(self):
      # Python 3
      if self.current >= self.vector.length:
        raise StopIteration
      obj = self.vector[self.current]
      self.current += 1
      return obj

    def next(self):
      # Python 2
      return self.__next__()

  def __len__(self):
    return self.length

  def __getitem__(self, i):
    try:
      # Python 2
      numeric_types = (int, long)
    except NameError:
      # Python 3
      numeric_types = int

    if isinstance(i, numeric_types):
      if i < 0:
        i += self.length
      if i > self.length:
        raise IndexError
      array_type = _Ptr(_Ptr(self._type_))
      return ctypes.cast(self.data, array_type)[i].contents
    return list(self)[i]

  def __iter__(self):
    return Vector.Iter(self)


Vector.EMPTY = Vector.in_dll(_dll, 'kGumboEmptyVector')


class AttributeVector(Vector):
  _type_ = Attribute


class NodeVector(Vector):
  # _type_ assigned later, to avoid circular references with Node
  pass


class QuirksMode(Enum):
  _values_ = ['NO_QUIRKS', 'QUIRKS', 'LIMITED_QUIRKS']


class Document(ctypes.Structure):
  _fields_ = [
      ('children', NodeVector),
      ('has_doctype', ctypes.c_bool),
      ('name', ctypes.c_char_p),
      ('public_identifier', ctypes.c_char_p),
      ('system_identifier', ctypes.c_char_p),
      ('doc_type_quirks_mode', QuirksMode),
      ]

  def __repr__(self):
    return 'Document'


class Namespace(Enum):
  URLS = [
      'http://www.w3.org/1999/xhtml',
      'http://www.w3.org/2000/svg',
      'http://www.w3.org/1998/Math/MathML',
  ]
  _values_ = ['HTML', 'SVG', 'MATHML']

  def to_url(self):
    return self.URLS[self.value]


class Tag(Enum):
  @staticmethod
  def from_str(tagname):
    text_ptr = ctypes.c_char_p(tagname.encode('utf-8'))
    return _tag_enum(text_ptr)

  _values_ = gumboc_tags.TagNames + ['UNKNOWN', 'LAST']

class Element(ctypes.Structure):
  _fields_ = [
      ('children', NodeVector),
      ('tag', Tag),
      ('tag_namespace', Namespace),
      ('original_tag', StringPiece),
      ('original_end_tag', StringPiece),
      ('start_pos', SourcePosition),
      ('end_pos', SourcePosition),
      ('attributes', AttributeVector),
      ]

  @property
  def tag_name(self):
    original_tag = StringPiece.from_buffer_copy(self.original_tag)
    _tag_from_original_text(ctypes.byref(original_tag))
    if self.tag_namespace == Namespace.SVG:
      svg_tagname = _normalize_svg_tagname(ctypes.byref(original_tag))
      if svg_tagname is not None:
        if PY3:
          return bytes(svg_tagname)
        return str(svg_tagname)
    if self.tag == Tag.UNKNOWN:
      if original_tag.data is None:
        return ''
      if PY3: 
          return (bytes(original_tag).decode('utf-8').lower()).encode('utf-8')
      return str(original_tag).lower()
    return _tagname(self.tag)

  def __repr__(self):
    return ('<%r>\n' % self.tag +
            '\n'.join(repr(child) for child in self.children) +
            '</%r>' % self.tag)


class Text(ctypes.Structure):
  _fields_ = [
      ('text', ctypes.c_char_p),
      ('original_text', StringPiece),
      ('start_pos', SourcePosition)
      ]

  def __repr__(self):
    return 'Text(%r)' % self.text

class OutputStatus(Enum):
  STATUS_MSG = [
      'OK',
      'Document tree depth limit exceeded',
      'System allocator returned NULL during parsing',
   ]
  _values_ = ['STATUS_OK', 'STATUS_TREE_TOO_DEEP', 'STATUS_OUT_OF_MEMORY']

  def to_string(self):
    return self.STATUS_MSG[self.value]


class NodeType(Enum):
  _values_ = ['DOCUMENT', 'ELEMENT', 'TEXT', 'CDATA',
              'COMMENT', 'WHITESPACE', 'TEMPLATE']


class NodeUnion(ctypes.Union):
  _fields_ = [
      ('document', Document),
      ('element', Element),
      ('text', Text),
      ]


class Node(ctypes.Structure):
  # _fields_ set later to avoid a circular reference

  def _contents(self):
    # Python3 enters an infinite loop if you use an @property within
    # __getattr__, so we factor it out to a helper.
    if self.type == NodeType.DOCUMENT:
      return self.v.document
    elif self.type in (NodeType.ELEMENT, NodeType.TEMPLATE):
      return self.v.element
    else:
      return self.v.text

  @property
  def contents(self):
    return self._contents()

  def __getattr__(self, name):
    return getattr(self._contents(), name)

  def __setattr__(self, name, value):
    return setattr(self._contents(), name, value)

  def __repr__(self):
    return repr(self.contents)


Node._fields_ = [
    ('type', NodeType),
    # Set the type to Node later to avoid a circular dependency.
    ('parent', _Ptr(Node)),
    ('index_within_parent', ctypes.c_uint),
    # TODO(jdtang): Make a real list of enum constants for this.
    ('parse_flags', _bitvector),
    ('v', NodeUnion)
    ]
NodeVector._type_ = Node


class Options(ctypes.Structure):
  _fields_ = [
      ('tab_stop', ctypes.c_int),
      ('use_xhtml_rules', ctypes.c_bool),
      ('stop_on_first_error', ctypes.c_bool),
      ('max_tree_depth', ctypes.c_uint),
      ('max_errors', ctypes.c_int),
      ]


class Output(ctypes.Structure):
  _fields_ = [
      ('document', _Ptr(Node)),
      ('root', _Ptr(Node)),
      ('status', OutputStatus),
      # TODO(jdtang): Error type.
      ('errors', Vector),
      ]

@contextlib.contextmanager
def parse(text, **kwargs):
  options = Options()
  context_tag = kwargs.get('container', Tag.LAST)
  context_namespace = kwargs.get('container_namespace', Namespace.HTML)
  for field_name, _ in Options._fields_:
    try:
      setattr(options, field_name, kwargs[field_name])
    except KeyError:
      setattr(options, field_name, getattr(_DEFAULT_OPTIONS, field_name))
  # We have to manually take a reference to the input text here so that it
  # outlives the parse output.  If we let ctypes do it automatically on function
  # call, it creates a temporary buffer which is destroyed when the call
  # completes, and then the original_text pointers point into invalid memory.
  # convert string to be utf-8 encoded
  if isinstance(text, text_type):
    text = text.encode('utf-8')
  text_ptr = ctypes.c_char_p(text)
  output = _parse_fragment(
      ctypes.byref(options), text_ptr, len(text),
      context_tag, context_namespace)
  try:
    yield output
  finally:
    _destroy_output(output)

_DEFAULT_OPTIONS = Options.in_dll(_dll, 'kGumboDefaultOptions')

_parse_with_options = _dll.gumbo_parse_with_options
_parse_with_options.argtypes = [_Ptr(Options), ctypes.c_char_p, ctypes.c_size_t]
_parse_with_options.restype = _Ptr(Output)

_parse_fragment = _dll.gumbo_parse_fragment
_parse_fragment.argtypes = [
    _Ptr(Options), ctypes.c_char_p, ctypes.c_size_t, Tag, Namespace]
_parse_fragment.restype = _Ptr(Output)

_tag_from_original_text = _dll.gumbo_tag_from_original_text
_tag_from_original_text.argtypes = [_Ptr(StringPiece)]
_tag_from_original_text.restype = None

_normalize_svg_tagname = _dll.gumbo_normalize_svg_tagname
_normalize_svg_tagname.argtypes = [_Ptr(StringPiece)]
_normalize_svg_tagname.restype = ctypes.c_char_p

_destroy_output = _dll.gumbo_destroy_output
_destroy_output.argtypes = [_Ptr(Output)]
_destroy_output.restype = None

_tagname = _dll.gumbo_normalized_tagname
_tagname.argtypes = [Tag]
_tagname.restype = ctypes.c_char_p

_tag_enum = _dll.gumbo_tag_enum
_tag_enum.argtypes = [ctypes.c_char_p]
_tag_enum.restype = Tag

__all__ = ['StringPiece', 'SourcePosition', 'AttributeNamespace', 'Attribute',
           'Vector', 'AttributeVector', 'NodeVector', 'QuirksMode', 'Document',
           'Namespace', 'Tag', 'Element', 'Text', 'NodeType', 'Node',
           'Options', 'Output', 'parse']



TagNames = [
	"A",
	"ABBR",
	"ACRONYM",
	"ADDRESS",
	"ALTGLYPH",
	"ALTGLYPHDEF",
	"ALTGLYPHITEM",
	"ANIMATE",
	"ANIMATECOLOR",
	"ANIMATEMOTION",
	"ANIMATETRANSFORM",
	"ANNOTATION_XML",
	"APPLET",
	"AREA",
	"ARTICLE",
	"ASIDE",
	"AUDIO",
	"B",
	"BASE",
	"BASEFONT",
	"BDI",
	"BDO",
	"BGSOUND",
	"BIG",
	"BLINK",
	"BLOCKQUOTE",
	"BODY",
	"BR",
	"BUTTON",
	"CANVAS",
	"CAPTION",
	"CENTER",
	"CIRCLE",
	"CITE",
	"CLIPPATH",
	"CODE",
	"COL",
	"COLGROUP",
	"COLOR_PROFILE",
	"CURSOR",
	"DATA",
	"DATALIST",
	"DD",
	"DEFS",
	"DEL",
	"DESC",
	"DETAILS",
	"DFN",
	"DIALOG",
	"DIR",
	"DIV",
	"DL",
	"DT",
	"ELLIPSE",
	"EM",
	"EMBED",
	"FEBLEND",
	"FECOLORMATRIX",
	"FECOMPONENTTRANSFER",
	"FECOMPOSITE",
	"FECONVOLVEMATRIX",
	"FEDIFFUSELIGHTING",
	"FEDISPLACEMENTMAP",
	"FEDISTANTLIGHT",
	"FEFLOOD",
	"FEFUNCA",
	"FEFUNCB",
	"FEFUNCG",
	"FEFUNCR",
	"FEGAUSSIANBLUR",
	"FEIMAGE",
	"FEMERGE",
	"FEMERGENODE",
	"FEMORPHOLOGY",
	"FEOFFSET",
	"FEPOINTLIGHT",
	"FESPECTACTUALRLIGHTING",
	"FESPOTLIGHT",
	"FETILE",
	"FETURBULENCE",
	"FIELDSET",
	"FIGCAPTION",
	"FIGURE",
	"FILTER",
	"FONT",
	"FONT_FACE",
	"FONT_FACE_FORMAT",
	"FONT_FACE_NAME",
	"FONT_FACE_SRC",
	"FONT_FACE_URI",
	"FOOTER",
	"FOREIGNOBJECT",
	"FORM",
	"FRAME",
	"FRAMESET",
	"G",
	"GLYPH",
	"GLYPHREF",
	"H1",
	"H2",
	"H3",
	"H4",
	"H5",
	"H6",
	"HEAD",
	"HEADER",
	"HGROUP",
	"HKERN",
	"HR",
	"HTML",
	"I",
	"IFRAME",
	"IMAGE",
	"IMG",
	"INPUT",
	"INS",
	"ISINDEX",
	"KBD",
	"KEYGEN",
	"LABEL",
	"LEGEND",
	"LI",
	"LINE",
	"LINEARGRADIENT",
	"LINK",
	"LISTING",
	"MACTION",
	"MAIN",
	"MALIGNGROUP",
	"MALIGNMARK",
	"MAP",
	"MARK",
	"MARKER",
	"MARQUEE",
	"MASK",
	"MATH",
	"MENCLOSE",
	"MENU",
	"MENUITEM",
	"MERROR",
	"META",
	"METADATA",
	"METER",
	"MFENCED",
	"MFRAC",
	"MGLYPH",
	"MI",
	"MISSING_GLYPH",
	"MLABELEDTR",
	"MLONGDIV",
	"MMULTISCRIPTS",
	"MN",
	"MO",
	"MOVER",
	"MPADDED",
	"MPATH",
	"MPHANTOM",
	"MPRESCRIPTS",
	"MROOT",
	"MROW",
	"MS",
	"MSCARRIES",
	"MSCARRY",
	"MSGROUP",
	"MSLINE",
	"MSPACE",
	"MSQRT",
	"MSROW",
	"MSTACK",
	"MSTYLE",
	"MSUB",
	"MSUBSUP",
	"MSUP",
	"MTABLE",
	"MTD",
	"MTEXT",
	"MTR",
	"MULTICOL",
	"MUNDER",
	"MUNDEROVER",
	"NAV",
	"NEXTID",
	"NOBR",
	"NOEMBED",
	"NOFRAMES",
	"NONE",
	"NOSCRIPT",
	"OBJECT",
	"OL",
	"OPTGROUP",
	"OPTION",
	"OUTPUT",
	"P",
	"PARAM",
	"PATH",
	"PATTERN",
	"PICTURE",
	"PLAINTEXT",
	"POLYGON",
	"POLYLINE",
	"PRE",
	"PROGRESS",
	"Q",
	"RADIALGRADIENT",
	"RB",
	"RECT",
	"RP",
	"RT",
	"RTC",
	"RUBY",
	"S",
	"SAMP",
	"SCRIPT",
	"SEARCH",
	"SECTION",
	"SELECT",
	"SEMANTICS",
	"SET",
	"SLOT",
	"SMALL",
	"SOURCE",
	"SPACER",
	"SPAN",
	"STOP",
	"STRIKE",
	"STRONG",
	"STYLE",
	"SUB",
	"SUMMARY",
	"SUP",
	"SVG",
	"SWITCH",
	"SYMBOL",
	"TABLE",
	"TBODY",
	"TD",
	"TEMPLATE",
	"TEXT",
	"TEXTPATH",
	"TEXTAREA",
	"TFOOT",
	"TH",
	"THEAD",
	"TIME",
	"TITLE",
	"TR",
	"TRACK",
	"TREF",
	"TSPAN",
	"TT",
	"U",
	"UL",
	"USE",
	"VAR",
	"VIDEO",
	"VIEW",
	"VKERN",
	"WBR",
	"XMP",
]



# Copyright 2012 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Adapter between Gumbo and BeautifulSoup.

This parses an HTML document and gives back a BeautifulSoup object, which you
can then manipulate like a normal BeautifulSoup parse tree.
"""

__author__ = 'jdtang@google.com (Jonathan Tang)'

import BeautifulSoup
import ctypes

import gumboc


def _utf8(text):
  return text.decode('utf-8', 'replace')


def _add_source_info(obj, original_text, start_pos, end_pos):
  obj.original = str(original_text)
  obj.line = start_pos.line
  obj.col = start_pos.column
  obj.offset = start_pos.offset
  obj.end_line = end_pos.line
  obj.end_col = end_pos.column
  obj.end_offset = end_pos.offset


def _add_document(element):
  # Currently ignored, since there's no real place for this in the BeautifulSoup
  # API.
  pass


def _add_text(cls):
  return lambda element: cls(_utf8(element.text))


def _convert_attrs(attrs):
  # TODO(jdtang): Ideally attributes would pass along their positions as well,
  # but I can't extend the built in str objects with new attributes.  Maybe work
  # around this with a subclass in some way...
  return [(_utf8(attr.name), _utf8(attr.value)) for attr in attrs]


class _Converter(object):
  def __init__(self, text, **kwargs):
    # We need to record the addresses of GumboNodes as we add them and correlate
    # them with the BeautifulSoup objects that they become.  This lets us
    # correctly wire up the next/previous pointers so that they point to
    # BeautifulSoup objects instead of ctypes ones.
    self._node_map = {}
    self._HANDLERS = [
        _add_document,
        self._add_element,
        _add_text(BeautifulSoup.NavigableString),
        _add_text(BeautifulSoup.CData),
        _add_text(BeautifulSoup.Comment),
        _add_text(BeautifulSoup.NavigableString),
        ]
    self.soup = BeautifulSoup.BeautifulSoup()
    with gumboc.parse(text, **kwargs) as output:
      self.soup.append(self._add_node(output.contents.root.contents))
    
    self._fix_next_prev_pointers(self.soup)

  def _add_element(self, element):
    tag = BeautifulSoup.Tag(
        self.soup, _utf8(element.tag_name), _convert_attrs(element.attributes))
    for child in element.children:
      tag.append(self._add_node(child))
    _add_source_info(
        tag, element.original_tag, element.start_pos, element.end_pos)
    tag.original_end_tag = str(element.original_end_tag)
    return tag

  def _add_node(self, node):
    result = self._HANDLERS[node.type.value](node.contents)

    try:
      result.next_addr = ctypes.addressof(node.next.contents)
    except (ValueError, AttributeError):
      # Null pointer.
      result.next_addr = 0

    try:
      result.prev_addr = ctypes.addressof(node.prev.contents)
    except (ValueError, AttributeError):
      # Null pointer.
      result.prev_addr = 0

    self._node_map[ctypes.addressof(node.contents)] = result
    return result

  def _fix_next_prev_pointers(self, tag):
    tag.next = self._node_map.get(tag.next_addr)
    tag.prev = self._node_map.get(tag.prev_addr)
    try:
      for child in tag.children:
        self._fix_next_prev_pointers(child)
    except (AttributeError, TypeError):
      # NavigableStrings
      pass


def parse(text, **kwargs):
  converter = _Converter(text, **kwargs)
  return converter.soup



#!/usr/bin/env python
# vim:fileencoding=utf-8
# License: Apache 2.0 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import re
import subprocess

self_path = os.path.abspath(__file__)
HEADER = '''\
// Do not edit
// Generated from {}.in (see gentags.py)

'''


def generate_tag_headers():
    with open("src/tag_strings.h", "wb") as tag_strings, \
            open("src/tag_enum.h", "wb") as tag_enum, \
            open("src/tag_sizes.h", "wb") as tag_sizes, \
            open("python/gumbo/gumboc_tags.py", "wb") as tag_py, \
            open('src/tag.in', 'rb') as tagfile:
        tag_py.write('TagNames = [\n'.encode('utf-8'))
        for f in (tag_strings, tag_enum, tag_sizes):
            f.write(HEADER.format('tag').encode('utf-8'))
        for tag in tagfile:
            tag = tag.decode('utf-8').strip()
            tag_upper = tag.upper().replace('-', '_')
            tag_py.write(('\t"%s",\n' % tag_upper).encode('utf-8'))
            tag_strings.write(('"%s",\n' % tag).encode('utf-8'))
            tag_enum.write(('GUMBO_TAG_%s,\n' % tag_upper).encode('utf-8'))
            tag_sizes.write(('%d, ' % len(tag)).encode('utf-8'))
            tag_sizes.write(b'\n')
        tag_py.write(']\n'.encode('utf-8'))

def generate_tag_perfect_hash(repetitions=200):
    raw = subprocess.check_output(
        'gperf -LANSI-C --ignore-case -H tag_hash -m{} src/tag.in'.format(repetitions).split()
    ).decode('utf-8').splitlines()
    for i, line in enumerate(raw):
        if line.startswith('in_word_set'):
            break
    else:
        raise SystemExit('Failed to find in_word_set()')
    lines = raw[:i - 1]
    del raw[:i - 1]
    raw = '\n'.join(raw)
    wordlist = re.search("wordlist\[\]\s+=\s+{(.*?)}", raw, re.DOTALL)
    if wordlist is None:
        raise SystemExit('Failed to find wordlist')
    wordlist = [w.strip().replace('"', '') for w in wordlist.group(1).split(',')]
    taglist = ["\tGUMBO_TAG_" + (w.upper().replace('-', '_') if w else 'LAST') for w in wordlist]
    processed = '\n'.join(lines) + '\n\n'
    processed += 'static const GumboTag kGumboTagMap[] = {\n%s\n};' % '\n,'.join(taglist)
    processed = re.sub(
        r'.+^tag_hash',
        HEADER.format('tag') + 'static inline unsigned int\ntag_hash',
        processed,
        flags=re.DOTALL | re.MULTILINE)
    with open('src/tag_perf.h', 'wb') as f:
        f.write(processed.encode('utf-8'))
        f.write(b'\n')


def main():
    os.chdir(os.path.dirname(self_path))
    generate_tag_headers()


if __name__ == '__main__':
    main()
    generate_tag_perfect_hash()



# Copyright 2012 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tests for the Gumbo's BeautifulSoup Python adapter."""

__author__ = 'jdtang@google.com (Jonathan Tang)'

import unittest

import soup_adapter


class SoupAdapterTest(unittest.TestCase):

  def testSimpleParse(self):
    soup = soup_adapter.parse(
        """
        <ul>
          <li class=odd><a href="one.html">One</a>
          <li class="even"><a href="two.html">Two</a>
          <li class='odd'><a href="three.html">Three</a>
          <li class="even"><a href="four.html">Four</a>
        </ul>
        """)

    head = soup.head
    self.assertEquals(soup, head.parent.parent)
    self.assertEquals(u'head', head.name)
    self.assertEquals(0, len(head))

    body = soup.body
    self.assertEquals(head, body.previousSibling)
    self.assertEquals(2, len(body))  # <ul> + trailing whitespace
    self.assertEquals(u'ul', body.contents[0].name)

    list_items = body.findAll('li')
    self.assertEquals(4, len(list_items))

    evens = body('li', 'even')
    self.assertEquals(2, len(evens))

    a2 = body.find('a', href='two.html')
    self.assertEquals(u'a', a2.name)
    self.assertEquals(u'Two', a2.contents[0])

    li2 = a2.parent
    self.assertEquals(u'li', li2.name)
    self.assertEquals(u'even', li2['class'])
    self.assertEquals(list_items[1], li2)
    self.assertEquals(evens[0], li2)

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab

from __future__ import unicode_literals, print_function

# Copyright 2015 Kevin B. Hendricks, Stratford, Ontario Canada
# Copyright 2012 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for Gumbo CTypes bindings."""

__author__ = 'jdtang@google.com (Jonathan Tang)'

import io
import unittest
import gumboc


class CtypesTest(unittest.TestCase):
    def testWordParse(self):
        with gumboc.parse('Test') as output:
            doctype_node = output.contents.document.contents
            self.assertEqual(gumboc.NodeType.DOCUMENT, doctype_node.type)
            document = doctype_node.v.document
            self.assertEqual(b'', document.name)
            self.assertEqual(b'', document.public_identifier)
            self.assertEqual(b'', document.system_identifier)

            root = output.contents.root.contents
            self.assertEqual(gumboc.NodeType.ELEMENT, root.type)
            self.assertEqual(gumboc.Tag.HTML, root.tag)
            self.assertEqual(gumboc.Namespace.HTML, root.tag_namespace)
            self.assertEqual(2, len(root.children))

            head = root.children[0]
            self.assertEqual(gumboc.NodeType.ELEMENT, head.type)
            self.assertEqual(gumboc.Tag.HEAD, head.tag)
            self.assertEqual(b'head', head.tag_name)
            self.assertEqual(gumboc.Namespace.HTML, head.tag_namespace)
            self.assertEqual(0, len(head.original_tag))
            self.assertEqual(b'', bytes(head.original_end_tag))
            self.assertEqual(0, head.children.length)

            body = root.children[1]
            self.assertNotEqual(body, doctype_node)
            self.assertEqual(gumboc.NodeType.ELEMENT, body.type)
            self.assertEqual(gumboc.Tag.BODY, body.tag)
            self.assertEqual(b'body', body.tag_name)
            self.assertEqual(1, len(body.children))

            text_node = body.children[0]
            self.assertEqual(gumboc.NodeType.TEXT, text_node.type)
            self.assertEqual(b'Test', text_node.text)

    def testBufferThatGoesAway(self):
        for i in range(10):
            source = io.StringIO('<foo bar=quux>1<p>2</foo>')
            parse_tree = gumboc.parse(source.read())
            source.close()
        with parse_tree as output:
            root = output.contents.root.contents
            body = root.children[1]
            foo = body.children[0]
            self.assertEqual(gumboc.NodeType.ELEMENT, foo.type)
            self.assertEqual(gumboc.Tag.UNKNOWN, foo.tag)
            self.assertEqual(b'<foo bar=quux>', bytes(foo.original_tag))
            self.assertEqual(b'', bytes(foo.original_end_tag))
            self.assertEqual(b'foo', foo.tag_name)
            self.assertEqual(b'bar', foo.attributes[0].name)
            self.assertEqual(b'quux', foo.attributes[0].value)

    def testUnknownTag(self):
        with gumboc.parse('<foo bar=quux>1<p>2</foo>') as output:
            root = output.contents.root.contents
            body = root.children[1]
            foo = body.children[0]
            self.assertEqual(gumboc.NodeType.ELEMENT, foo.type)
            self.assertEqual(gumboc.Tag.UNKNOWN, foo.tag)
            self.assertEqual(b'<foo bar=quux>', bytes(foo.original_tag))
            self.assertEqual(b'', bytes(foo.original_end_tag))
            self.assertEqual(b'foo', foo.tag_name)
            self.assertEqual(b'bar', foo.attributes[0].name)
            self.assertEqual(b'quux', foo.attributes[0].value)

    def testSarcasm(self):
        with gumboc.parse('<div><sarcasm><div></div></sarcasm></div>') as output:
            root = output.contents.root.contents
            body = root.children[1]
            div = body.children[0]
            sarcasm = div.children[0]
            self.assertEqual(gumboc.NodeType.ELEMENT, sarcasm.type)
            self.assertEqual(gumboc.Tag.UNKNOWN, sarcasm.tag)
            self.assertEqual(b'<sarcasm>', bytes(sarcasm.original_tag))
            self.assertEqual(b'</sarcasm>', bytes(sarcasm.original_end_tag))
            self.assertEqual(b'sarcasm', sarcasm.tag_name)

    def testEnums(self):
        self.assertEqual(gumboc.Tag.A, gumboc.Tag.A)
        self.assertEqual(hash(gumboc.Tag.A.value), hash(gumboc.Tag.A))

    def testFragment(self):
        with gumboc.parse(
            '<div></div>',
            container=gumboc.Tag.TITLE,
            container_namespace=gumboc.Namespace.SVG) as output:
            root = output.contents.root.contents
            self.assertEqual(1, len(root.children))
            div = root.children[0]
            self.assertEqual(gumboc.NodeType.ELEMENT, div.type)
            self.assertEqual(gumboc.Tag.DIV, div.tag)
            self.assertEqual(gumboc.Namespace.HTML, div.tag_namespace)

if __name__ == '__main__':
    unittest.main()



# Copyright 2012 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Adapter between Gumbo and html5lib.

This exports one method, parse, with the same signature as html5lib.parse.  It
takes the text to parse, and optionally an html5lib TreeBuilder to build the
tree, and gives back a DOM tree in that format.  Example:

  doc = parse(text, treebuilder='lxml')
"""

__author__ = 'jdtang@google.com (Jonathan Tang)'

import gumboc

# These should match html5lib.constants.namespaces, and be indexed by the enum
# values of gumboc.Namespace
_NAMESPACES = [
    'http://www.w3.org/1999/xhtml',
    'http://www.w3.org/2000/svg',
    'http://www.w3.org/1998/Math/MathML',
    ]


def _convert_doctype(treebuilder, source_node):
  if not source_node.has_doctype:
    # Mimic html5lib behavior: if no doctype token, no doctype node.
    return
  treebuilder.insertDoctype({
      'name': source_node.name.decode('utf-8'),
      'publicId': source_node.public_identifier.decode('utf-8'),
      'systemId': source_node.system_identifier.decode('utf-8'),
      })


def _convert_attributes(source_node):
  def maybe_namespace(attr):
    if attr.namespace != gumboc.AttributeNamespace.NONE:
      return (repr(attr.namespace).lower() if attr.name != 'xmlns' else None,
              attr.name.decode('utf-8'),
              attr.namespace.to_url())
    else:
      return attr.name.decode('utf-8')
  return dict((maybe_namespace(attr), attr.value.decode('utf-8'))
              for attr in source_node.attributes)


def _convert_element(source_node):
  if source_node.type not in ( gumboc.NodeType.ELEMENT, gumboc.NodeType.TEMPLATE):
    # If-statement instead of assert so it runs with -O
    raise AssertionError(
        '_convert_element only works with elements; found %r' %
        source_node.type)
  return {
      'name': source_node.v.element.tag_name.decode('utf-8'),
      'namespace': _NAMESPACES[source_node.v.element.tag_namespace.value],
      'data': _convert_attributes(source_node),
      }


def _insert_root(treebuilder, source_node, pop_element = True):
  treebuilder.insertRoot(_convert_element(source_node))
  for child_node in source_node.children:
    _insert_node(treebuilder, child_node)
  if pop_element:
    treebuilder.openElements.pop()

def _insert_node(treebuilder, source_node):
  assert source_node.type != gumboc.NodeType.DOCUMENT
  if source_node.type == gumboc.NodeType.COMMENT:
    treebuilder.insertComment({'data': source_node.v.text.text.decode('utf-8')})
  elif source_node.type in (
      gumboc.NodeType.TEXT,
      gumboc.NodeType.WHITESPACE,
      gumboc.NodeType.CDATA):
    treebuilder.insertText(source_node.v.text.text.decode('utf-8'))
  else:
    treebuilder.insertElementNormal(_convert_element(source_node))
    for child_node in source_node.v.element.children:
      _insert_node(treebuilder, child_node)
    treebuilder.openElements.pop()


class HTMLParser(object):
  def __init__(self, tree):
    self.tree = tree

  def parse(self, text_or_file, **kwargs):
    # need to tell sigil gumbo to use html parsing rules not xhtml parsing rules
    kwargs["use_xhtml_rules"] = False
    try:
      text = text_or_file.read()
    except AttributeError:
      # Assume a string.
      text = text_or_file
    with gumboc.parse(text, **kwargs) as output:
      _convert_doctype(self.tree, output.contents.document.contents)
      for node in output.contents.document.contents.children:
        if node.type == gumboc.NodeType.COMMENT:
          self.tree.insertComment({'data': node.v.text.text.decode('utf-8')},
                                  self.tree.document)
        elif node.type in (gumboc.NodeType.ELEMENT, gumboc.NodeType.TEMPLATE):
          _insert_root(self.tree, output.contents.root.contents)
        else:
          assert 'Only comments and <html> nodes allowed at the root'
      return self.tree.getDocument()

  def parseFragment(self, text_or_file, container, **kwargs):
    # need to tell sigil gumbo to use html parsing rules not xhtml parsing rules
    kwargs["use_xhtml_rules"] = False
    try:
      text = text_or_file.read()
    except AttributeError:
      # Assume a string.
      text = text_or_file
    if ' ' in container:
      container_ns, container = container.split(' ')
    else:
      container_ns = "html"

    with gumboc.parse(
        text,
        container=gumboc.Tag.from_str(container),
        container_namespace=getattr(gumboc.Namespace, container_ns.upper()),
        **kwargs) as output:
      for node in output.contents.document.contents.children:
        if node.type in (gumboc.NodeType.ELEMENT, gumboc.NodeType.TEMPLATE):
          _insert_root(self.tree, output.contents.root.contents, False)
        else:
          assert 'Malformed fragment parse (??)'
      return self.tree.getFragment()



# Copyright 2012 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tests for the Gumbo => Html5lib adapter."""
import codecs
import collections
import glob
import os
import re
import io
    
import unittest
import warnings

from html5lib import treebuilders

import html5lib_adapter


TREEBUILDER = treebuilders.getTreeBuilder('dom')
TESTDATA_BASE_PATH = os.path.join(
    os.path.split(__file__)[0], '..', '..')


# Copied from html5lib.tests/test_parser.py
def convertTreeDump(data):
  return "\n".join(convertExpected(data, 3).split("\n")[1:])


# Copied/adapted/simplified from html5lib.tests/support.py
def html5lib_test_files():
  return glob.glob(os.path.join(
      TESTDATA_BASE_PATH, 'testdata', 'tree-construction', '*.dat'))


class TestData(object):
  def __init__(self, filename):
    self.f = codecs.open(filename, encoding="utf8")

  def __iter__(self):
    data = collections.defaultdict(lambda: None)
    key=None
    for line in self.f:
      heading = self.isSectionHeading(line)
      if heading:
        if data and heading == 'data':
          #Remove trailing newline
          data[key] = data[key][:-1]
          yield self.normaliseOutput(data)
          data = collections.defaultdict(lambda: None)
        key = heading
        data[key] = ''
      elif key is not None:
          data[key] += line
    if data:
      yield self.normaliseOutput(data)

  def isSectionHeading(self, line):
    """If the current heading is a test section heading return the heading,
    otherwise return False"""
    if line.startswith("#"):
      return line[1:].strip()
    else:
      return False

  def normaliseOutput(self, data):
    # Remove trailing newlines
    for key, value in data.items():
      if value.endswith("\n"):
        data[key] = value[:-1]
    return data

def convertExpected(data, stripChars):
  """convert the output of str(document) to the format used in the testcases"""
  data = data.split("\n")
  rv = []
  for line in data:
    if line.startswith("|"):
      rv.append(line[stripChars:])
    else:
      rv.append(line)
  return "\n".join(rv)

def reformatTemplateContents(expected):
 lines = expected.split('\n')
 retval = []
 template_indents = []
 for line in lines:
   line_stripped = line.strip()
   indent = len(line) - len(line_stripped)
   if line_stripped == 'content':
     template_indents.append(indent)
     continue
   while template_indents and indent <= template_indents[-1]:
       template_indents.pop()
   if template_indents:
     line = line[2 * len(template_indents):]
   retval.append(line)
 return '\n'.join(retval)

class Html5libAdapterTest(unittest.TestCase):
  """Adapter between Gumbo and the html5lib tests.

  This works through a bit of magic.  It's an empty class at first, but then
  buildTestCases runs through the test files in html5lib, and adds a
  method to this class for each one.  That method acts like
  test_parser.TestCase.runParserTest, running a parse, serializing the tree, and
  comparing it to the expected output.

  The vague name is so nosetests doesn't try to run it as a test.
  """
  def impl(self, inner_html, input, expected, errors):
    p = html5lib_adapter.HTMLParser(
            tree=TREEBUILDER(namespaceHTMLElements=True))

    if inner_html:
      document = p.parseFragment(
          io.StringIO(input), inner_html.replace('math ', 'mathml '))
    else:
      document = p.parse(io.StringIO(input))

    with warnings.catch_warnings():
      # Etree serializer in html5lib uses a deprecated getchildren() API.
      warnings.filterwarnings('ignore', category=DeprecationWarning)
      output = convertTreeDump(p.tree.testSerializer(document))

    expected = re.compile(r'^(\s*)<(\S+)>', re.M).sub(
        r'\1<html \2>', convertExpected(expected, 2))
    # html5lib doesn't yet support the template tag, but it appears in the
    # tests with the expectation that the template contents will be under the
    # word 'contents', so we need to reformat that string a bit.
    expected = reformatTemplateContents(expected)

    error_msg = '\n'.join(['\n\nInput:', input, '\nExpected:', expected,
                           '\nReceived:', output])
    self.assertEqual(expected, output,
                      error_msg.encode('ascii', 'xmlcharrefreplace') + b'\n')
    # TODO(jdtang): Check error messages, when there's full error support.


def BuildTestCases(cls):
  for filename in html5lib_test_files():
    test_name = os.path.basename(filename).replace('.dat', '')
    for i, test in enumerate(TestData(filename)):
      # html5lib parses <noscript> tags as if the scripting-enabled flag is
      # set, while we parse as if the scripting-disabled flag is set (since we
      # don't really support scripting and the resulting parse tree is often
      # more useful for toolsmiths).  That means our output will differ by
      # design from html5lib's, so we disable any of their tests that involve
      # <noscript>
      if '<noscript>' in test['data']:
        continue

      # <command> has been renamed to <menuitem> in recent versions of the spec.
      # html5lib 0.95 does not include this yet, and so we disable tests that
      # include the old tag.
      if '<command>' in test['data']:
        continue

      def test_func(
          self,
          inner_html=test['document-fragment'],
          input=test['data'],
          expected=test['document'],
          errors=test.get('errors', '').split('\n')):
        return self.impl(inner_html, input, expected, errors)
      test_func.__name__ = 'test_%s_%d' % (test_name, i + 1)
      setattr(cls, test_func.__name__, test_func)


if __name__ == '__main__':
  BuildTestCases(Html5libAdapterTest)
  unittest.main()



# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab

from __future__ import unicode_literals, print_function

# Copyright 2012 Google Inc. All Rights Reserved.
# Modifications to use BeautifulSoup4 
#   Copyright 2015 Kevin B. Hendricks, Stratford, Ontario, Canada
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# Should this be reworked to be a bs4 treebuilder?

"""
  Adapter between Gumbo and BeautifulSoup4.
  This parses an HTML document and gives back a BeautifulSoup4 object, which you
  can then manipulate like a normal BeautifulSoup4 parse tree.

  Groks namespaces on elements and attributes
"""

__author__ = 'jdtang@google.com (Jonathan Tang)'

import sys
import gumboc

import bs4
# uses bs4.element classes:
#      Comment, DocType, NavigableString, CData, Tag, NamespacedAttribute, whitespace_re

# These should be indexed by the enum
# values of gumboc.Namespace

_NAMESPACES = [
    'http://www.w3.org/1999/xhtml',
    'http://www.w3.org/2000/svg',
    'http://www.w3.org/1998/Math/MathML',
    ]


def _fromutf8(text):
    return text.decode('utf-8', 'replace')


def _add_source_info(obj, original_text, start_pos, end_pos):
    obj.original = _fromutf8(bytes(original_text))
    obj.line = start_pos.line
    obj.col = start_pos.column
    obj.offset = start_pos.offset
    if end_pos:
        obj.end_line = end_pos.line
        obj.end_col = end_pos.column
        obj.end_offset = end_pos.offset


def _convert_attrs(element_attrs):
    def maybe_namespace(attr):
        if attr.namespace != gumboc.AttributeNamespace.NONE:
            name = _fromutf8(attr.name)
            prefix = repr(attr.namespace).lower() if name != 'xmlns' else None
            nsurl = attr.namespace.to_url()
            return bs4.element.NamespacedAttribute(prefix, name, nsurl)
        else:
            return _fromutf8(attr.name)
    def maybe_value_list(attr):
        value = _fromutf8(attr.value)
        if " " in value:
            value = bs4.element.whitespace_re.split(value)
        return value
    return dict((maybe_namespace(attr), maybe_value_list(attr)) for attr in element_attrs)


def _add_document(soup, element):
    if not element.has_doctype:
        # Mimic html5lib behavior: if no doctype token, no doctype node.
        return
    doctype = bs4.element.Doctype.for_name_and_ids(_fromutf8(element.name),
                                                   _fromutf8(element.public_identifier),
                                                   _fromutf8(element.system_identifier))
    soup.object_was_parsed(doctype)


def _add_element(soup, element):
    tag = bs4.element.Tag(parser=soup,
                  name=_fromutf8(element.tag_name),
                  namespace=_NAMESPACES[element.tag_namespace.value],
                  attrs=_convert_attrs(element.attributes))
    for child in element.children:
        tag.append(_add_node(soup, child))
    _add_source_info(tag, element.original_tag, element.start_pos, element.end_pos)
    tag.original_end_tag = _fromutf8(bytes(element.original_end_tag))
    return tag


def _add_text(cls):
    def add_text_internal(soup, element):
        text = cls(_fromutf8(element.text))
        _add_source_info(text, element.original_text, element.start_pos, None)
        return text
    return add_text_internal


_HANDLERS = [
    _add_document,                              # DOCUMENT
    _add_element,                               # ELEMENT
    _add_text(bs4.element.NavigableString),     # TEXT
    _add_text(bs4.element.CData),               # CDATA
    _add_text(bs4.element.Comment),             # COMMENT
    _add_text(bs4.element.NavigableString),     # WHITESPACE
    _add_element,                               # TEMPLATE
    ]


def _add_node(soup, node):
    return _HANDLERS[node.type.value](soup, node.contents)


def _add_next_prev_pointers(soup):
    def _traverse(node):
        # .findAll requires the .next pointer, which is what we're trying to add
        # when we call this, and so we manually supply a generator to yield the
        # nodes in DOM order.
        yield node
        try:
            for child in node.contents:
                for descendant in _traverse(child):
                    yield descendant
        except AttributeError:
            # Not an element.
            return
    nodes = sorted(_traverse(soup), key=lambda node: node.offset)
    if nodes:
        nodes[0].previous_element = None
        nodes[-1].next_element = None
    for i, node in enumerate(nodes[1:-1], 1):
        nodes[i-1].next_element = node
        node.previous_element = nodes[i-1]


def parse(text, **kwargs):
    with gumboc.parse(text, **kwargs) as output:
        soup = bs4.BeautifulSoup('', "html.parser")
        _add_document(soup, output.contents.document.contents)
        for node in output.contents.document.contents.children:
            soup.append(_add_node(soup, node))
        _add_next_prev_pointers(soup.html)
        return soup


def main():
    samp = """
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"
  "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml/" xml:lang="en" lang="en-US">
<head><title>testing & entities</title></head>
<body>
  <p class="first second">this&nbsp;is&#160;the&#xa0;<i><b>copyright</i></b> symbol "&copy;"</p>
  <p xmlns:xlink="http://www.w3.org/xlink" class="second" xlink:href="http://www.ggogle.com">this used to test atribute namespaces</p>
</body>
</html>
"""
    soup = parse(samp)
    print(soup.decode())
    for node in soup.findAll("head"):
        print(node)
    for node in soup.find_all(attrs={'class':'second'}):
        print(node)
    return 0

if __name__ == '__main__':
    sys.exit(main())
